from sm.namespaces.namespace import (
    DefaultKnowledgeGraphNamespace,
    KnowledgeGraphNamespace,
    Namespace,
    OutOfNamespace,
)
from sm.namespaces.utils import KGName, get_kgns, has_kgns, register_kgns
from sm.namespaces.wikidata import WikidataNamespace

__all__ = [
    "Namespace",
    "OutOfNamespace",
    "WikidataNamespace",
    "KnowledgeGraphNamespace",
    "DefaultKnowledgeGraphNamespace",
    "KGName",
    "get_kgns",
    "register_kgns",
    "has_kgns",
]



from __future__ import annotations

from curses.ascii import isdigit
from enum import Enum
from typing import Optional

from sm.namespaces.dbpedia import DBpediaNamespace
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.namespaces.wikidata import ExtendedWikidataNamespace
from sm.typing import InternalID


class KGName(str, Enum):
    Wikidata = "wikidata"
    DBpedia = "dbpedia"
    Generic = "generic"

    def __str__(self):
        return self.value


registered_kgns: dict[str, KnowledgeGraphNamespace] = {
    KGName.Wikidata: ExtendedWikidataNamespace.create(),
    KGName.DBpedia: DBpediaNamespace.create(),
}


def get_kgns(kgname: KGName) -> KnowledgeGraphNamespace:
    if kgname in registered_kgns:
        return registered_kgns[kgname]
    raise NotImplementedError(kgname)


def register_kgns(kgname: str, kgns: KnowledgeGraphNamespace):
    global registered_kgns
    registered_kgns[kgname] = kgns


def has_kgns(kgname: str) -> bool:
    global registered_kgns
    return kgname in registered_kgns



from sm.outputs._sm_formatter import deser_simple_tree_yaml, ser_simple_tree_yaml
from sm.outputs._sm_transform import (
    create_sm_from_cta_cpa,
    create_sm_nodes,
    remove_isolated_nodes,
    remove_literal_nodes,
    replace_class_nodes_by_subject_columns,
)
from sm.outputs.semantic_model import (
    ClassNode,
    DataNode,
    Edge,
    LiteralNode,
    LiteralNodeDataType,
    SemanticModel,
)

__all__ = [
    "SemanticModel",
    "DataNode",
    "ClassNode",
    "LiteralNode",
    "Edge",
    "LiteralNodeDataType",
    "replace_class_nodes_by_subject_columns",
    "remove_isolated_nodes",
    "create_sm_from_cta_cpa",
    "create_sm_nodes",
    "ser_simple_tree_yaml",
    "deser_simple_tree_yaml",
    "remove_literal_nodes",
]



from __future__ import annotations

import csv
import random
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from hashlib import md5
from io import BytesIO, StringIO
from operator import attrgetter
from pathlib import Path
from typing import (
    Generator,
    Generic,
    Literal,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
)
from urllib.parse import urlparse
from zipfile import Path as ZipPath
from zipfile import ZipFile

import orjson
import pandas as pd
import serde.csv
import serde.yaml
from ruamel.yaml import YAML
from serde import json
from serde.helper import DEFAULT_ORJSON_OPTS, get_open_fn
from slugify import slugify
from sm.inputs.prelude import ColumnBasedTable, Context, Link
from sm.misc.funcs import batch
from sm.misc.matrix import Matrix
from sm.namespaces.namespace import Namespace
from sm.outputs import deser_simple_tree_yaml, ser_simple_tree_yaml
from sm.outputs.semantic_model import SemanticModel
from tqdm.auto import tqdm
from typing_extensions import Self

T = TypeVar("T", covariant=True)
T1 = TypeVar("T1")


@dataclass
class Example(Generic[T]):
    id: str
    sms: list[SemanticModel]
    table: T

    def replace_table(self, table: T1) -> Example[T1]:
        return Example(id=self.id, sms=self.sms, table=table)


@dataclass
class FullTable:
    table: ColumnBasedTable
    context: Context
    links: Matrix[list[Link]]

    def nrows(self) -> int:
        return self.table.nrows()

    def ncols(self) -> int:
        return self.table.ncols()

    def sample_rows(self, num_rows: int) -> FullTable:
        """Sample a subset of rows"""
        if self.nrows() > num_rows:
            indices = random.sample(range(self.nrows()), num_rows)
            return self.select_rows(indices)
        return self

    def select_rows(self, indices: list[int]) -> FullTable:
        """Select a subset of rows based on a boolean mask"""
        return FullTable(
            table=self.table.select_rows(indices),
            context=self.context,
            links=Matrix([self.links.data[i] for i in indices]),
        )

    def keep_columns(self, columns: list[int], reindex: bool = False) -> FullTable:
        """Keep only the specified columns"""
        if reindex:
            links = Matrix([[row[ci] for ci in columns] for row in self.links.data])
        else:
            # if not re-indexing the columns, we need to keep the original
            # shape (non-selected columns will be empty lists)
            ignore_cols = [
                col.index for col in self.table.columns if col.index not in columns
            ]
            links = self.links.shallow_copy()
            for row in links.data:
                for ci in ignore_cols:
                    row[ci] = []
        return FullTable(
            table=self.table.keep_columns(columns, reindex),
            context=self.context,
            links=links,
        )

    def remove_empty_links(self) -> Self:
        return self.__class__(
            table=self.table,
            context=self.context,
            links=self.links.map(
                lambda links: [link for link in links if link.end > link.start]
            ),
        )

    def to_dict(self):
        return {
            "version": 2,
            "table": self.table.to_dict(),
            "context": self.context.to_dict(),
            "links": [
                [[link.to_dict() for link in cell] for cell in row]
                for row in self.links.data
            ],
        }

    @classmethod
    def from_dict(cls, obj: dict):
        version = obj["version"]
        if not (version == "1.2" or version == "1.1" or version == 2):
            raise ValueError(f"Unknown version: {version}")

        table = ColumnBasedTable.from_dict(obj["table"])
        if "links" in obj:
            links = Matrix(
                [
                    [[Link.from_dict(link) for link in cell] for cell in row]
                    for row in obj["links"]
                ]
            )
        else:
            links = Matrix.default(table.shape(), list)

        return cls(
            table=table,
            context=Context.from_dict(obj["context"]),
            links=links,
        )

    @classmethod
    def from_column_based_table(cls, table: ColumnBasedTable) -> FullTable:
        context = Context()
        links = Matrix.default(table.shape(), list)
        return cls(table=table, context=context, links=links)


@dataclass
class Dataset:
    location: Path

    @contextmanager
    def _open(self) -> Generator[Union[ZipPath, Path], None, None]:
        if self.is_zip_file():
            # it is a zip file
            with ZipFile(self.location, mode="r") as zf:
                root = ZipPath(zf)
                if self.description_dir(root).exists():
                    yield root
                else:
                    subdirs = list(root.iterdir())
                    if len(subdirs) == 1 and self.description_dir(subdirs[0]).exists():
                        yield subdirs[0]
                    else:
                        raise ValueError("Invalid dataset format")
        else:
            yield self.location

    def description_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "descriptions"

    def table_dir(self, root: Union[Path, ZipPath]) -> Union[Path, ZipPath]:
        return root / "tables"

    def is_zip_file(self):
        return self.location.name.endswith(".zip")

    def load(self, verbose: bool = False) -> list[Example[FullTable]]:
        """Load dataset from a folder or a single zip file. Assuming the following structure:

        descriptions (containing semantic descriptions of tables)
        ├── <table_fs_id>
        │   ├── version.01[.json|.yml|.st.yml]
        │   ├── version.02[.json|.yml|.st.yml]
        │   └── ...
            or
        ├── <table_fs_id>[.json|.yml|.st.yml]
        ├── ...
        tables (containing list of tables, the type of table depends on )
        ├── <table_fs_id>[.json|.csv|.xlsx][.gz|.bz2|.lz4]
        ├── ...

        We also support compressing formats such as .zip.
        descriptions
        ├── part-<num>.zip (no nested version folders)
        │   ├── <table_fs_id>[.json|.yml|.st.yml]
        |   |   ...
        tables
        ├── part-<num>.zip
        │   ├── <table_fs_id>[.json|.csv|.xlsx]
        |   |   ...

        The description can be specified in either json or yaml format. For more information on
        how the semantic models are deserialized from the format, checkout the corresponding
        deserialization functions.

        Args:
            data_dir:

        Returns:
        """
        examples = []
        with self._open() as root:
            descdir = self.description_dir(root)
            tabledir = self.table_dir(root)

            for infile in tqdm(
                sorted(tabledir.iterdir(), key=attrgetter("name")), disable=not verbose
            ):
                suffixes = Path(infile.name).suffixes
                if infile.name.startswith(".") or len(suffixes) == 0:
                    continue

                if suffixes[-1] != ".zip":
                    example_id = infile.name[: -sum(len(x) for x in suffixes)]
                    table = self._deser_table(
                        example_id, infile.read_bytes(), suffixes[0]
                    )

                    if descdir.exists():
                        if (descdir / example_id).exists():
                            desc_file = max(
                                (descdir / example_id).iterdir(),
                                key=lambda file: int(file.name.split(".")[1]),
                            )
                            assert desc_file is not None
                        else:
                            desc_file = descdir / f"{example_id}.json"
                            if not desc_file.exists():
                                desc_file = descdir / f"{example_id}.yml"
                                if not desc_file.exists():
                                    desc_file = descdir / f"{example_id}.st.yml"
                            assert (
                                desc_file.exists()
                            ), f"Description file not found for {example_id}"

                        sms = self._deser_sm(
                            table.table,
                            desc_file.read_bytes(),
                            "".join(desc_file.suffixes),
                        )
                    else:
                        sms = []

                    examples.append(
                        Example(id=table.table.table_id, sms=sms, table=table)
                    )
                else:
                    assert (
                        infile.name.endswith(".zip")
                        and isinstance(infile, Path)
                        and isinstance(descdir, Path)
                        and not self.is_zip_file()
                    ), "Must not be a zip file"
                    part: dict[str, FullTable] = {}
                    with ZipFile(infile, mode="r") as zf:
                        for file in zf.infolist():
                            if not file.filename.endswith(".json"):
                                continue

                            table_id = Path(file.filename).stem
                            with zf.open(file, mode="r") as f:
                                table = self._deser_table(
                                    table_id, f.read(), Path(file.filename).suffixes[0]
                                )
                            part[table_id] = table

                    if descdir.exists():
                        lst = []
                        with ZipFile(descdir / infile.name, mode="r") as zf:
                            for file in zf.infolist():
                                table_id = Path(file.filename).stem
                                if table_id not in part:
                                    continue

                                with zf.open(file, mode="r") as f:
                                    sms = self._deser_sm(
                                        part[table_id].table,
                                        f.read(),
                                        "".join(file.filename.split(".")[1:]),
                                    )
                                    lst.append(
                                        Example(
                                            id=part[table_id].table.table_id,
                                            sms=sms,
                                            table=part[table_id],
                                        )
                                    )
                    else:
                        lst = [
                            Example(id=table.table.table_id, sms=[], table=table)
                            for table in part.values()
                        ]
                    assert len(lst) == len(part)
                    examples.extend(lst)

            return examples

    def save(
        self,
        examples: list[Example[FullTable]],
        individual_table_compressed: Optional[Literal["gz", "bz2", "lz4"]] = None,
        batch_compressed: bool = False,
        batch_size: int = 100,
        sm_fmt: Literal["json", "simple-tree"] = "json",
        sm_fmt_indent: Literal[0, 2] = 0,
        table_fmt: Literal["json", "txt"] = "json",
        table_fmt_indent: Literal[0, 2] = 0,
        clean_previous_data: bool = True,
        ns: Optional[Namespace] = None,
        multi_desc_version: bool = False,
    ):
        """Save dataset to a folder, single, or multiple zip files. Assuming the following structure:

        Args:
            multiversion: if True, we create a description folder for each example id so that we can have multiple versioned files
        """
        if self.is_zip_file():
            with ZipFile(self.location, "w") as root:
                assert isinstance(root, ZipFile)
                for e in examples:
                    ename = get_friendly_fs_id(e.table.table.table_id)
                    root.writestr(
                        "descriptions/" + ename + f".{fmt_to_ext(sm_fmt)}",
                        data=self._ser_sm(
                            e.sms, e.table.table, sm_fmt, sm_fmt_indent, ns
                        ),
                        # data=orjson.dumps([sm.to_dict() for sm in e.sms]),
                    )
                    root.writestr(
                        "tables/" + ename + f".{fmt_to_ext(table_fmt)}",
                        data=self._ser_table(e.table, table_fmt, table_fmt_indent),
                    )
            return

        descdir = self.description_dir(self.location)
        tabledir = self.table_dir(self.location)
        assert (
            not self.is_zip_file()
            and isinstance(descdir, Path)
            and isinstance(tabledir, Path)
        )

        if descdir.exists() and clean_previous_data:
            shutil.rmtree(descdir)
        descdir.mkdir(parents=True, exist_ok=True)

        if tabledir.exists() and clean_previous_data:
            shutil.rmtree(tabledir)
        tabledir.mkdir(parents=True, exist_ok=True)

        if batch_compressed:
            for i, bexamples in enumerate(batch(batch_size, examples)):
                bexamples: list[Example[FullTable]]
                filename = f"part-{i:04d}.zip"
                with ZipFile(descdir / filename, "w") as dzf, ZipFile(
                    tabledir / filename, "w"
                ) as tzf:
                    for e in bexamples:
                        ename = get_friendly_fs_id(e.table.table.table_id)
                        dzf.writestr(
                            ename + f".{fmt_to_ext(sm_fmt)}",
                            data=self._ser_sm(
                                e.sms, e.table.table, sm_fmt, sm_fmt_indent, ns
                            ),
                        )
                        tzf.writestr(
                            ename + f".{fmt_to_ext(table_fmt)}",
                            data=self._ser_table(e.table, table_fmt, table_fmt_indent),
                        )
        else:
            for e in examples:
                filename = get_friendly_fs_id(e.table.table.table_id)

                if multi_desc_version:
                    desc_outfile = (
                        descdir / filename / f"version.01.{fmt_to_ext(sm_fmt)}"
                    )
                    desc_outfile.parent.mkdir(parents=True, exist_ok=True)
                else:

                    desc_outfile = descdir / f"{filename}.{fmt_to_ext(sm_fmt)}"
                desc_outfile.write_bytes(
                    self._ser_sm(e.sms, e.table.table, sm_fmt, sm_fmt_indent, ns)
                )

                table_outfile = tabledir / (
                    filename + f".{fmt_to_ext(table_fmt)}.{individual_table_compressed}"
                    if individual_table_compressed is not None
                    else filename + f".{fmt_to_ext(table_fmt)}"
                )
                with get_open_fn(table_outfile)(table_outfile, "wb") as f:
                    f.write(self._ser_table(e.table, table_fmt, table_fmt_indent))

    @staticmethod
    def _deser_table(table_id: str, data: bytes, ext: str) -> FullTable:
        if ext == ".json":
            return FullTable.from_dict(orjson.loads(data))
        if ext == ".csv":
            column_based_table = ColumnBasedTable.from_dataframe(
                pd.read_csv(BytesIO(data)),
                table_id=table_id,
            )
            return FullTable(
                table=column_based_table,
                context=Context(),
                links=Matrix.default(column_based_table.shape(), list),
            )
        if ext == ".xlsx":
            column_based_table = ColumnBasedTable.from_dataframe(
                pd.read_excel(BytesIO(data)),
                table_id=table_id,
            )
            return FullTable(
                table=column_based_table,
                context=Context(),
                links=Matrix.default(column_based_table.shape(), list),
            )
        if ext == ".txt":
            part1_csv, part2_json = data.split(b"\n" + b"-" * 80 + b"\n\n")
            column_based_table = ColumnBasedTable.from_dataframe(
                pd.read_csv(BytesIO(part1_csv)),
                table_id=table_id,
            )
            obj = orjson.loads(part2_json)
            return FullTable(
                table=column_based_table,
                context=Context.from_dict(obj["context"]),
                links=Matrix(
                    [
                        [[Link.from_dict(link) for link in cell] for cell in row]
                        for row in obj["links"]
                    ]
                ),
            )

        raise ValueError(f"Unknown file type: {ext}")

    @staticmethod
    def _ser_table(
        table: FullTable,
        table_fmt: Literal["json", "txt"] = "json",
        table_fmt_indent: Literal[0, 2] = 0,
    ):
        if table_fmt == "json":
            if table_fmt_indent > 0:
                orjson_opts = DEFAULT_ORJSON_OPTS | orjson.OPT_INDENT_2
            else:
                orjson_opts = DEFAULT_ORJSON_OPTS

            return orjson.dumps(table.to_dict(), option=orjson_opts)

        if table_fmt == "txt":
            out = StringIO()
            nrows, ncols = table.table.shape()

            writer = csv.writer(
                out, delimiter=",", quoting=csv.QUOTE_MINIMAL, lineterminator="\n"
            )
            writer.writerow([c.name for c in table.table.columns])
            for ri in range(nrows):
                writer.writerow(
                    [table.table.columns[ci].values[ri] for ci in range(ncols)]
                )

            out.write("\n" + "-" * 80 + "\n\n")
            if table_fmt_indent > 0:
                orjson_opts = DEFAULT_ORJSON_OPTS | orjson.OPT_INDENT_2
            else:
                orjson_opts = DEFAULT_ORJSON_OPTS

            out.write(
                orjson.dumps(
                    {
                        "context": table.context.to_dict(),
                        "links": [
                            [[link.to_dict() for link in cell] for cell in row]
                            for row in table.links.data
                        ],
                    },
                    option=orjson_opts,
                ).decode()
            )
            return out.getvalue().encode()

        raise ValueError(f"Unsupported format: {table_fmt}")

    @staticmethod
    def _deser_sm(table: ColumnBasedTable, data: bytes, ext: str):
        if ext == ".json":
            return [SemanticModel.from_dict(sm) for sm in orjson.loads(data)]
        if ext == ".st.yml":
            return [deser_simple_tree_yaml(table, BytesIO(data))]
        if ext == ".yml":
            yaml = YAML()
            raw = yaml.load(BytesIO(data))
            ns = Namespace.from_prefix2ns(raw["prefixes"])
            sms = [SemanticModel.from_yaml_dict(sm, ns) for sm in raw["models"]]
            return sms
        raise ValueError(f"Unknown file type: {ext}")

    @staticmethod
    def _ser_sm(
        sm: Union[SemanticModel, Sequence[SemanticModel]],
        table: ColumnBasedTable,
        fmt: Literal["json", "simple-tree"],
        fmt_indent: Literal[0, 2] = 0,
        ns: Optional[Namespace] = None,
    ):
        if fmt == "json":
            if isinstance(sm, SemanticModel):
                sm = [sm]
            return orjson.dumps(
                [x.to_dict() for x in sm],
                option=orjson.OPT_INDENT_2 if fmt_indent > 0 else None,
            )
        elif fmt == "simple-tree":
            if isinstance(sm, Sequence):
                assert len(sm) == 1
                sm = sm[0]

            assert ns is not None
            file = BytesIO()

            ser_simple_tree_yaml(table, sm, ns, file)
            return file.getvalue()
        else:
            raise ValueError(f"Unsupported format: {fmt}")


def get_friendly_fs_id(id: str) -> str:
    if id.startswith("http://") or id.startswith("https://"):
        if id.find("dbpedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        if id.find("wikipedia.org") != -1:
            return (
                slugify(
                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                ).replace("-", "_")
                + "_"
                + md5(id.encode()).hexdigest()
            )

        raise NotImplementedError()
    return slugify(id.replace("/", "_"), lowercase=False).replace("-", "_")


def fmt_to_ext(fmt: Literal["json", "txt", "simple-tree"]) -> str:
    if fmt == "simple-tree":
        return "st.yml"
    return fmt


class SampleableTable(Protocol):
    def nrows(self) -> int: ...

    def select_rows(self, indices: list[int]) -> Self: ...


ST = TypeVar("ST", bound=SampleableTable, covariant=True)


def sample_table_data(
    examples: Sequence[Example[ST]], n_rows: int, seed: Optional[int] = None
) -> Sequence[Example[ST]]:
    """Sample data from each table in examples.

    Args:
        examples: list of examples
        n_rows: number of rows to sample per table
        seed: random seed
    """
    rng = random.Random(seed)
    new_exs = []
    for ex in examples:
        tbl_nrows = ex.table.nrows()
        if tbl_nrows <= n_rows:
            new_exs.append(ex)
        else:
            indices = rng.sample(range(tbl_nrows), n_rows)
            new_exs.append(
                Example(id=ex.id, sms=ex.sms, table=ex.table.select_rows(indices))
            )

    return new_exs



from dataclasses import dataclass
from typing import Dict, Mapping, Optional, Sequence, Set, Union

from sm.evaluation import sm_metrics
from sm.evaluation.utils import PrecisionRecallF1
from sm.outputs import (
    ClassNode,
    DataNode,
    SemanticModel,
    remove_isolated_nodes,
    replace_class_nodes_by_subject_columns,
)


@dataclass
class CTAEvalOutput(PrecisionRecallF1):
    n_corrects: float  # float as we allow for partial correctness
    n_examples: int
    n_predictions: int


def cpa(
    gold_sm: SemanticModel,
    pred_sm: SemanticModel,
    id_props: Set[str],
    scoring_fn: Optional[sm_metrics.ScoringFn] = None,
) -> sm_metrics.SmPrecisionRecallF1Output:
    gold_sm = gold_sm.deep_copy()
    pred_sm = pred_sm.deep_copy()

    _cpa_transformation(gold_sm, id_props)
    _cpa_transformation(pred_sm, id_props)

    return sm_metrics.precision_recall_f1(
        gold_sm=gold_sm, pred_sm=pred_sm, scoring_fn=scoring_fn
    )


def cta(
    gold_sm: SemanticModel,
    pred_sm: SemanticModel,
    id_props: Set[str],
    scoring_fn: Optional[sm_metrics.ScoringFn] = None,
    ignored_columns: Optional[Set[str]] = None,
) -> CTAEvalOutput:
    gold_cta = _get_cta(gold_sm, id_props)
    pred_cta = _get_cta(pred_sm, id_props)

    if ignored_columns is not None:
        gold_cta = {k: v for k, v in gold_cta.items() if k not in ignored_columns}
        pred_cta = {k: v for k, v in pred_cta.items() if k not in ignored_columns}

    if scoring_fn is None:
        scoring_fn = sm_metrics.ScoringFn()

    return _cta_real(gold_cta, pred_cta, scoring_fn)


def _cta_real(
    gold_cta: Mapping[str, Union[Sequence[str], Set[str], str]],
    pred_cta: Mapping[str, str],
    scoring_fn: sm_metrics.ScoringFn,
):
    score = 0.0
    for cindex in set(gold_cta.keys()).intersection(pred_cta.keys()):
        gc = gold_cta[cindex]
        pc = pred_cta[cindex]

        if not isinstance(gc, str):
            score += max(scoring_fn.get_match_score(pc, g) for g in gc)
        else:
            score += scoring_fn.get_match_score(pc, gc)

    if len(pred_cta) == 0:
        precision = 1.0
    else:
        precision = score / len(pred_cta)

    if len(gold_cta) == 0:
        recall = 1.0
    else:
        recall = score / len(gold_cta)

    if recall + precision == 0:
        f1 = 0
    else:
        f1 = 2 * precision * recall / (precision + recall)
    return CTAEvalOutput(
        precision=precision,
        recall=recall,
        f1=f1,
        n_corrects=score,
        n_examples=len(gold_cta),
        n_predictions=len(pred_cta),
    )


def _get_cta(sm: SemanticModel, id_props: Set[str]) -> Dict[str, str]:
    col2class = {}
    for n in sm.iter_nodes():
        if isinstance(n, ClassNode):
            outedges = sm.out_edges(n.id)
            id_edges = [outedge for outedge in outedges if outedge.abs_uri in id_props]
            if len(id_edges) > 1:
                raise Exception("Haven't supported multiple subject columns yet")
            if len(id_edges) == 0:
                continue
            dnode = sm.get_node(id_edges[0].target)
            assert isinstance(dnode, DataNode)
            col2class[str(dnode.col_index)] = n.abs_uri
    return col2class


def _cpa_transformation(sm: SemanticModel, id_props: Set[str]) -> None:
    replace_class_nodes_by_subject_columns(sm, id_props)
    remove_isolated_nodes(sm)



from __future__ import annotations

import typing
from collections import Counter
from pathlib import Path
from typing import Sequence

import serde.yaml
from graph.retworkx import has_cycle
from sm.inputs.table import ColumnBasedTable
from sm.misc.prelude import UnreachableError
from sm.namespaces.prelude import Namespace
from sm.outputs.semantic_model import (
    ClassNode,
    DataNode,
    Edge,
    LiteralNode,
    LiteralNodeDataType,
    SemanticModel,
)


def ser_simple_tree_yaml(
    table: ColumnBasedTable, sm: SemanticModel, ns: Namespace, outfile: Path | typing.IO
):
    """Save the semantic model to a YAML file with simple formatting as follow:

    ```yaml
    version: "simple-tree-1"
    model:
        - type: <class_uri>
          props:
            <property_uri>: <node>

    prefixes:
        <namespace>: <url>
    ```

    If `<node>` can be:
        - the column name (string)
        - column index (`{column: <index>}`)
        - a literal value (`{literal:<type>: <value>}`)
        - an entity value (`{entity: <value>}`)
        - a nested class (`{type: <class_uri>, props: {...}}`)
        - a list of the above for multiple values for the same property.

    Note:
        - the model must be a tree, an error will be thrown if the model contains cycles.
        - columns that are not used in the model will not be included in the output.

    Args:
        table: The table containing the columns referenced in the semantic model
        sm: The semantic model to serialize
        ns: The namespace used for URI resolution
        outfile: Path or file-like object where to write the serialized output
    """
    if has_cycle(sm):
        raise ValueError(
            "The model contains cycles, cannot convert to the simple YAML format"
        )

    output = {"version": "simple-tree-1", "model": [], "prefixes": {}}
    used_uris = set()

    # precompute the column format
    col_fmt: dict[int, str | dict] = {}
    counter = Counter((col.clean_name for col in table.columns))
    for col in table.columns:
        colname = col.clean_name or ""
        if colname.strip() == "" or counter[col.clean_name] > 1:
            col_fmt[col.index] = {"column": col.index}
        else:
            col_fmt[col.index] = colname

    def serialize_node(node: ClassNode | LiteralNode | DataNode):
        if isinstance(node, ClassNode):
            used_uris.add(node.abs_uri)
            outdict = {"type": ns.get_rel_uri(node.abs_uri), "props": {}}
            for edge in sm.out_edges(node.id):
                used_uris.add(edge.abs_uri)
                prop = ns.get_rel_uri(edge.abs_uri)
                ser_node = serialize_node(sm.get_node(edge.target))
                if prop in outdict["props"]:
                    if not isinstance(outdict["props"], Sequence):
                        outdict["props"][prop] = [outdict["props"][prop]]
                    outdict["props"][prop].append(ser_node)
                else:
                    outdict["props"][prop] = ser_node
            return outdict

        if isinstance(node, DataNode):
            return col_fmt[node.col_index]
        if isinstance(node, LiteralNode):
            if node.datatype == LiteralNodeDataType.Entity:
                key = "entity"
                outdict = {key: node.value, "props": {}}
                for edge in sm.out_edges(node.id):
                    used_uris.add(edge.abs_uri)
                    prop = ns.get_rel_uri(edge.abs_uri)
                    ser_node = serialize_node(sm.get_node(edge.target))
                    if prop in outdict["props"]:
                        if not isinstance(outdict["props"], Sequence):
                            outdict["props"][prop] = [outdict["props"][prop]]
                        outdict["props"][prop].append(ser_node)
                    else:
                        outdict["props"][prop] = ser_node
                if len(outdict["props"]) == 0:
                    del outdict["props"]
            else:
                key = "literal-{}".format(node.datatype.value)
                outdict = {key: node.value}

            return outdict

        raise UnreachableError()

    for u in sm.iter_nodes():
        if sm.in_degree(u.id) > 0:
            continue
        if isinstance(u, DataNode):
            # skip data nodes that are not used
            continue
        output["model"].append(serialize_node(u))

    for uri in used_uris:
        prefix = ns.prefix_index.get(uri)
        if prefix is not None:
            output["prefixes"][prefix] = ns.prefix2ns[prefix]

    return serde.yaml.ser(output, outfile)


def deser_simple_tree_yaml(
    table: ColumnBasedTable, infile: Path | typing.IO
) -> SemanticModel:
    indict = serde.yaml.deser(infile)
    assert indict["version"] == "simple-tree-1"

    sm = SemanticModel()
    name2col = {col.clean_name: col for col in table.columns}
    if len(indict["prefixes"]) == 0:
        assert len(indict["model"]) == 0
        return sm
    namespace = Namespace.from_prefix2ns(indict["prefixes"])

    def deserialize_node(
        obj: list | dict | str | int,
    ) -> DataNode | LiteralNode | ClassNode | list[DataNode | LiteralNode | ClassNode]:
        if isinstance(obj, str):
            if not sm.has_data_node(name2col[obj].index):
                # must be a column name
                node = DataNode(col_index=name2col[obj].index, label=obj)
                sm.add_node(node)
            else:
                node = sm.get_data_node(name2col[obj].index)
            return node

        if isinstance(obj, list):
            # recursive deserialization
            return [deserialize_node(o) for o in obj]  # type: ignore

        if isinstance(obj, int):
            node = DataNode(
                col_index=obj,
                label=table.get_column_by_index(obj).clean_name or "",
            )
            sm.add_node(node)
            return node

        assert isinstance(obj, dict)
        if "column" in obj:
            node = DataNode(
                col_index=obj["column"],
                label=table.get_column_by_index(obj["column"]).clean_name or "",
            )
            sm.add_node(node)
            return node

        if "type" in obj:
            node = ClassNode(
                abs_uri=namespace.get_abs_uri(obj["type"]), rel_uri=obj["type"]
            )
            sm.add_node(node)
            for prop, value in obj["props"].items():
                deser_values = deserialize_node(value)
                if not isinstance(deser_values, list):
                    deser_values = [deser_values]
                for deser_value in deser_values:
                    edge = Edge(
                        source=node.id,
                        target=deser_value.id,
                        abs_uri=namespace.get_abs_uri(prop),
                        rel_uri=prop,
                    )
                    sm.add_edge(edge)
            return node

        if "entity" in obj:
            node = LiteralNode(
                value=obj["entity"],
                datatype=LiteralNodeDataType.Entity,
            )
            if sm.has_literal_node(node.value):
                return sm.get_literal_node(node.value)
            sm.add_node(node)

            for prop, value in obj.get("props", {}).items():
                deser_values = deserialize_node(value)
                if not isinstance(deser_values, list):
                    deser_values = [deser_values]
                for deser_value in deser_values:
                    edge = Edge(
                        source=node.id,
                        target=deser_value.id,
                        abs_uri=namespace.get_abs_uri(prop),
                        rel_uri=prop,
                    )
                    sm.add_edge(edge)
            return node

        _type, _value = next(iter((obj.items())))
        assert _type.startswith("literal-")
        node = LiteralNode(
            value=_value,
            datatype=LiteralNodeDataType(_type.split("-", 1)[1]),
        )
        if sm.has_literal_node(node.value):
            return sm.get_literal_node(node.value)
        sm.add_node(node)
        return node

    for node in indict["model"]:
        deserialize_node(node)
    return sm



#!/usr/bin/python
# -*- coding: utf-8 -*-

import math
import os
from dataclasses import dataclass
from itertools import chain, permutations
from typing import Callable, Dict, Generator, List, Optional, Sequence, Set, Tuple

from loguru import logger
from pyrsistent import PVector, pvector
from sm.evaluation.utils import PrecisionRecallF1
from sm.outputs.semantic_model import ClassNode, DataNode, LiteralNode, SemanticModel

"""
Compute precision, recall and f1 score of the semantic model according to Mohsen paper.

Convention: x' and x are nodes in predicted model and gold model, respectively.
"""


class PermutationExplosion(Exception):
    pass


class Node(object):
    __slots__ = ("id", "label", "incoming_links", "outgoing_links")

    def __init__(self, id: int, label: str) -> None:
        self.id: int = id
        self.label: str = label
        self.incoming_links: List[Link] = []
        self.outgoing_links: List[Link] = []

    @staticmethod
    def add_incoming_link(node: "Node", link: "Link"):
        node.incoming_links.append(link)
        link.target = node

    @staticmethod
    def add_outgoing_link(node: "Node", link: "Link"):
        node.outgoing_links.append(link)
        link.source = node

    def __str__(self):
        return "Node(id=%s, label=%s)" % (self.id, self.label)


@dataclass(frozen=True, eq=True)
class NodeTriple:
    __slots__ = ("source_id", "link_label", "target_id")
    source_id: int
    link_label: str
    target_id: int

    def __setstate__(self, state):
        assert state[0] is None and isinstance(state[1], dict)
        for slot, value in state[1].items():
            object.__setattr__(self, slot, value)


class Link(object):
    __slots__ = ("id", "label", "source", "source_id", "target", "target_id")

    def __init__(self, id: int, label: str, source: Node, target: Node) -> None:
        self.id: int = id
        self.label: str = label
        self.source = source
        self.source_id: int = source.id
        self.target = target
        self.target_id: int = target.id


class LabelGroup(object):
    """Represent a group of nodes that have same label"""

    __slots__ = ("nodes", "node_triples", "size")

    def __init__(self, nodes: List[Node]) -> None:
        self.nodes: List[Node] = nodes
        self.node_triples: Set[NodeTriple] = {
            NodeTriple(link.source_id, link.label, link.target_id)
            for node in self.nodes
            for link in chain(node.incoming_links, node.outgoing_links)
        }
        self.size: int = len(nodes)

    def __repr__(self):
        return "(#nodes=%d)" % (len(self.nodes))

    # noinspection PyUnusedLocal
    @staticmethod
    def group_by_structures(group: "LabelGroup", pred_group: "LabelGroup"):
        """
        A structure of a node is defined by its links, or we can treat it as a set of triple.
        Unbounded nodes should be assumed to be different, therefore a node have unbounded nodes will have
        it own structure group.
        We need not consider triple that are impossible to map to node in pred_group. This trick will improve
        the performance.
        """
        # TODO: implement it
        return [StructureGroup([n]) for n in group.nodes]


class StructureGroup(object):
    """Represent a group of nodes that have same structure"""

    __slots__ = ("nodes", "size")

    def __init__(self, nodes: List[Node]) -> None:
        self.nodes: List[Node] = nodes
        self.size: int = len(nodes)


class PairLabelGroup(object):
    __slots__ = ("X", "X_prime", "label")

    def __init__(self, label: str, X: LabelGroup, X_prime: LabelGroup) -> None:
        self.label = label
        self.X: LabelGroup = X
        self.X_prime: LabelGroup = X_prime

    def __repr__(self):
        return "(label=%s, X=%s, X_prime=%s)" % (self.label, self.X, self.X_prime)


class DependentGroups(object):
    """Represent a list of groups of nodes that are dependent on each other"""

    __slots__ = ("pair_groups", "X_triples", "X_prime_triples", "x_pairs")

    def __init__(self, pair_groups: List[PairLabelGroup]):
        self.pair_groups: List[PairLabelGroup] = pair_groups
        self.X_triples: Set[NodeTriple] = set()
        self.X_prime_triples: Set[NodeTriple] = set()

        for pair in pair_groups:
            self.X_triples = self.X_triples.union(pair.X.node_triples)
            self.X_prime_triples = self.X_prime_triples.union(pair.X_prime.node_triples)

        # a mapping from (source id, target id) to list of predicates
        # use this for computing if we want to take into account the ancestor/descendant in the predicates
        self.x_pairs = {}
        for triple in self.X_triples:
            if (triple.source_id, triple.target_id) not in self.x_pairs:
                self.x_pairs[triple.source_id, triple.target_id] = []
            self.x_pairs[triple.source_id, triple.target_id].append(triple.link_label)

    def get_n_permutations(self):
        n_permutation = 1
        for pair_group in self.pair_groups:
            n = max(pair_group.X.size, pair_group.X_prime.size)
            m = min(pair_group.X.size, pair_group.X_prime.size)
            n_permutation *= math.factorial(n) / math.factorial(n - m)

        return n_permutation


class PartialBijection(object):
    """A bijection defined a one-one mapping from x' => x.
    A node is mapped to None means that it doesn't have a correspondent node in the other model.
    """

    __slots__ = ("prime2x", "x2prime")

    def __init__(self) -> None:
        # a map from x' => x (pred_sm to gold_sm)
        self.prime2x: Dict[int, Optional[int]] = {}
        # map from x => x'
        self.x2prime: Dict[int, Optional[int]] = {}

    @staticmethod
    def construct_from_mapping(
        mapping: Sequence[Tuple[Optional[int], Optional[int]]]
    ) -> "PartialBijection":
        """
        :param mapping: a list of map from x' => x
        """
        self = PartialBijection()
        self.prime2x = {x_prime: x for x_prime, x in mapping if x_prime is not None}
        self.x2prime = {x: x_prime for x_prime, x in mapping if x is not None}
        return self

    def extends(self, bijection: "PartialBijection") -> "PartialBijection":
        another = PartialBijection()
        another.prime2x = dict(self.prime2x)
        another.prime2x.update(bijection.prime2x)
        another.x2prime = dict(self.x2prime)
        another.x2prime.update(bijection.x2prime)
        return another

    def extends_(self, bijection: "PartialBijection") -> None:
        self.prime2x.update(bijection.prime2x)
        self.x2prime.update(bijection.x2prime)

    def is_gold_node_bounded(self, node_id: int) -> bool:
        return node_id in self.x2prime

    def is_pred_node_bounded(self, node_id: int) -> bool:
        return node_id in self.prime2x


class IterGroupMapsUsingGroupingArgs(object):
    __slots__ = ("node_index", "bijection", "G_sizes")

    def __init__(self, node_index: int, bijection: PVector, G_sizes) -> None:
        self.node_index: int = node_index
        self.bijection: PVector = bijection
        self.G_sizes = G_sizes


class IterGroupMapsGeneralApproachArgs(object):
    __slots__ = ("node_index", "bijection")

    def __init__(self, node_index: int, bijection: PVector) -> None:
        self.node_index: int = node_index
        self.bijection: PVector = bijection


class FindBestMapArgs(object):
    __slots__ = ("group_index", "bijection")

    def __init__(self, group_index: int, bijection: PartialBijection) -> None:
        self.group_index: int = group_index
        self.bijection = bijection


class ScoringFn:
    # noinspection PyMethodMayBeStatic
    def get_match_score(self, pred_predicate: str, target_predicate: str) -> float:
        return int(pred_predicate == target_predicate)


@dataclass
class SmPrecisionRecallF1Output(PrecisionRecallF1):
    bijection: PartialBijection
    n_corrects: float  # float as we allow for partial correctness
    n_examples: int
    n_predictions: int
    gold_triples: Set[NodeTriple]
    pred_triples: Set[NodeTriple]


def find_best_map(
    dependent_group: DependentGroups, bijection: PartialBijection, scoring_fn: ScoringFn
) -> PartialBijection:
    terminate_index: int = len(dependent_group.pair_groups)
    # This code find the size of this
    # array: sum([min(gold_group.size, pred_group.size) for gold_group, pred_group in dependent_group.groups])
    call_stack = [FindBestMapArgs(group_index=0, bijection=bijection)]
    n_called = 0

    best_map = None
    best_score = -1

    while True:
        call_args = call_stack.pop()
        n_called += 1
        if call_args.group_index == terminate_index:
            # it is terminated, calculate score
            score = eval_score(dependent_group, call_args.bijection, scoring_fn)
            if score > best_score:
                best_score = score
                best_map = call_args.bijection
        else:
            pair_group = dependent_group.pair_groups[call_args.group_index]
            X, X_prime = pair_group.X, pair_group.X_prime
            for group_map in iter_group_maps(X, X_prime, call_args.bijection):
                bijection = PartialBijection.construct_from_mapping(group_map)

                call_stack.append(
                    FindBestMapArgs(
                        group_index=call_args.group_index + 1,
                        bijection=call_args.bijection.extends(bijection),
                    )
                )

        if len(call_stack) == 0:
            break

    return best_map  # type: ignore


def get_unbounded_nodes(
    X: LabelGroup, is_bounded_func: Callable[[int], bool]
) -> List[Node]:
    """Get nodes of a label group which have not been bounded by a bijection"""
    unbounded_nodes = []

    for x in X.nodes:
        for link in x.incoming_links:
            if not is_bounded_func(link.source_id):
                unbounded_nodes.append(link.source)

        for link in x.outgoing_links:
            if not is_bounded_func(link.target_id):
                unbounded_nodes.append(link.target)

    return unbounded_nodes


def get_common_unbounded_nodes(
    X: LabelGroup, X_prime: LabelGroup, bijection: PartialBijection
) -> Set[str]:
    """Finding unbounded nodes in X and X_prime that have same labels"""
    unbounded_X = get_unbounded_nodes(X, bijection.is_gold_node_bounded)
    unbounded_X_prime = get_unbounded_nodes(X_prime, bijection.is_pred_node_bounded)

    labeled_unbounded_X = {}
    labeled_unbounded_X_prime = {}
    for x in unbounded_X:
        if x.label not in labeled_unbounded_X:
            labeled_unbounded_X[x.label] = []
        labeled_unbounded_X[x.label].append(x)

    for x in unbounded_X_prime:
        if x.label not in labeled_unbounded_X_prime:
            labeled_unbounded_X_prime[x.label] = []
        labeled_unbounded_X_prime[x.label].append(x)

    common_unbounded_nodes = set(labeled_unbounded_X.keys()).intersection(
        labeled_unbounded_X_prime.keys()
    )
    return common_unbounded_nodes


def group_dependent_elements(dependency_map: List[List[int]]) -> List[int]:
    # algorithm to merge the dependencies
    # input:
    #   - dependency_map: [<element_index, ...>, ...] list of dependencies where element at ith position is list of index of elements
    #           that element at ith position depends upon.
    # output:
    #   - dependency_groups: [<group_id>, ...] list of group id, where element at ith position is group id that element belongs to
    dependency_groups: List[int] = [-1 for _ in range(len(dependency_map))]
    invert_dependency_groups = {}

    for i, g in enumerate(dependency_map):
        dependent_elements = g + [i]
        groups = {dependency_groups[j] for j in dependent_elements}
        valid_groups = groups.difference([-1])
        if len(valid_groups) == 0:
            group_id = len(invert_dependency_groups)
            invert_dependency_groups[group_id] = set()
        else:
            group_id = next(iter(valid_groups))

        if -1 in groups:
            # map unbounded elements to group has group_id
            for j in dependent_elements:
                if dependency_groups[j] == -1:
                    dependency_groups[j] = group_id
                    invert_dependency_groups[group_id].add(j)

        for another_group_id in valid_groups.difference([group_id]):
            for j in invert_dependency_groups[another_group_id]:
                dependency_groups[j] = group_id
                invert_dependency_groups[group_id].add(j)

    return dependency_groups


def split_by_dependency(
    map_groups: List[PairLabelGroup], bijection: PartialBijection
) -> List[DependentGroups]:
    """This method takes a list of groups (X, X') and group them based on their dependencies.
    D = {D1, D2, …} s.t for all Di, Dj, (Xi, Xi') in Di, (Xj, Xj’) in Dj, they are independent

    Two groups of nodes are dependent when at least one unbounded nodes in a group is a label of other group.
    For example, "actor_appellation" has link to "type", so group "actor_appellation" depends on group "type"
    """
    group_label2idx = {map_group.label: i for i, map_group in enumerate(map_groups)}

    # build group dependencies
    dependency_map = [[] for _ in range(len(map_groups))]
    for i, map_group in enumerate(map_groups):
        X, X_prime = map_group.X, map_group.X_prime
        common_labels = get_common_unbounded_nodes(X, X_prime, bijection)

        for common_label in common_labels:
            group_id = group_label2idx[common_label]
            dependency_map[i].append(group_id)

    dependency_groups = group_dependent_elements(dependency_map)
    dependency_pair_groups: Dict[int, List[PairLabelGroup]] = {}
    dependency_map_groups: List[DependentGroups] = []

    for i, map_group in enumerate(map_groups):
        if dependency_groups[i] not in dependency_pair_groups:
            dependency_pair_groups[dependency_groups[i]] = []
        dependency_pair_groups[dependency_groups[i]].append(map_group)

    for pair_groups in dependency_pair_groups.values():
        dependency_map_groups.append(DependentGroups(pair_groups))

    return dependency_map_groups


def iter_group_maps(
    X: LabelGroup, X_prime: LabelGroup, bijection: PartialBijection
) -> Generator[List[Tuple[int, Optional[int]]], None, None]:
    """Generate all mapping between X and X_prime. Return a list of tuple of (x_prime, x)"""
    if X.size < X_prime.size:
        return iter_group_maps_general_approach(X, X_prime)
    else:
        G = LabelGroup.group_by_structures(X, X_prime)
        return iter_group_maps_using_grouping(X_prime, G)


def iter_group_maps_general_approach(
    X: LabelGroup, X_prime: LabelGroup
) -> Generator[List[Tuple[int, Optional[int]]], None, None]:
    """
    Generate all mapping from X to X_prime
    NOTE: |X| < |X_prime|

    Return mapping from (x_prime to x)
    """
    mapping_mold: List[Optional[int]] = [None for _ in range(X_prime.size)]

    for perm in permutations(range(X_prime.size), X.size):
        mapping: List[Tuple[int, Optional[int]]] = []
        for i, j in enumerate(perm):
            mapping_mold[j] = X.nodes[i].id

        for i in range(X_prime.size):
            mapping.append((X_prime.nodes[i].id, mapping_mold[i]))
            mapping_mold[i] = None
        yield mapping


def iter_group_maps_using_grouping(
    X_prime: LabelGroup, G: List[StructureGroup]
) -> Generator[List[Tuple[int, Optional[int]]], None, None]:
    """
    Generate all mapping from X_prime to G (nodes in X grouped by their structures)
    NOTE: |X_prime| <= |X|

    Return mapping from (x_prime to x)
    """
    G_sizes = pvector((g.size for g in G))
    bijection: PVector = pvector([-1 for _ in range(X_prime.size)])
    terminate_index: int = X_prime.size

    call_stack: List[IterGroupMapsUsingGroupingArgs] = [
        IterGroupMapsUsingGroupingArgs(
            node_index=0, bijection=bijection, G_sizes=G_sizes
        )
    ]

    while True:
        call_args = call_stack.pop()

        if call_args.node_index == terminate_index:
            # convert bijection into a final mapping
            G_numerator = [0 for _ in range(len(G))]
            bijection = call_args.bijection
            mapping = []
            for i in range(len(bijection)):
                x_prime = X_prime.nodes[i].id
                x = G[bijection[i]].nodes[G_numerator[bijection[i]]].id

                G_numerator[bijection[i]] += 1
                mapping.append((x_prime, x))

            yield mapping
        else:
            for i, G_i in enumerate(G):
                if call_args.G_sizes[i] == 0:
                    continue

                bijection = call_args.bijection.set(call_args.node_index, i)
                G_sizes = call_args.G_sizes.set(i, call_args.G_sizes[i] - 1)

                call_stack.append(
                    IterGroupMapsUsingGroupingArgs(
                        node_index=call_args.node_index + 1,
                        bijection=bijection,
                        G_sizes=G_sizes,
                    )
                )

        if len(call_stack) == 0:
            break


def prepare_args(
    gold_sm: "SemanticModel", pred_sm: "SemanticModel"
) -> List[PairLabelGroup]:
    """Prepare data for evaluation

    + data_node_mode = 0, mean we don't touch anything (note that the label of data_node must be unique)
    + data_node_mode = 1, mean we ignore label of data node (convert it to DATA_NODE, DATA_NODE2 if there are duplication columns)
    + data_node_mode = 2, mean we ignore data node
    """

    def convert_graph(graph: "SemanticModel"):
        node_index: Dict[int, Node] = {}

        for v in graph.iter_nodes():
            if isinstance(v, ClassNode):
                label = v.abs_uri
            elif isinstance(v, DataNode):
                label = f"attr-{v.col_index:02d}"
            else:
                assert isinstance(v, LiteralNode)
                label = v.value
            node_index[v.id] = Node(v.id, label)

        for i, e in enumerate(graph.iter_edges()):
            link = Link(i, e.abs_uri, node_index[e.source], node_index[e.target])
            Node.add_outgoing_link(node_index[e.source], link)
            Node.add_incoming_link(node_index[e.target], link)

        return node_index

    label2nodes = {}
    gold_id2node = convert_graph(gold_sm)
    pred_id2node = convert_graph(pred_sm)

    for node in gold_id2node.values():
        if node.label not in label2nodes:
            label2nodes[node.label] = ([], [])
        label2nodes[node.label][0].append(node)
    for node in pred_id2node.values():
        if node.label not in label2nodes:
            label2nodes[node.label] = ([], [])
        label2nodes[node.label][1].append(node)

    return [
        PairLabelGroup(label, LabelGroup(g[0]), LabelGroup(g[1]))
        for label, g in label2nodes.items()
    ]


def eval_score(
    dependent_groups: DependentGroups,
    bijection: PartialBijection,
    scoring_fn: ScoringFn,
) -> float:
    x_pairs = dependent_groups.x_pairs
    mapped_xprime_triples = {}
    for triple in dependent_groups.X_prime_triples:
        if (
            triple.source_id not in bijection.prime2x
            or triple.target_id not in bijection.prime2x
        ):
            continue

        s = bijection.prime2x[triple.source_id]
        o = bijection.prime2x[triple.target_id]
        if (s, o) in x_pairs:
            if (s, o) not in mapped_xprime_triples:
                mapped_xprime_triples[(s, o)] = []
            mapped_xprime_triples[(s, o)].append(triple.link_label)

    score = 0.0
    for so, edges in mapped_xprime_triples.items():
        if len(x_pairs[so]) == 1:
            gold_edge = x_pairs[so][0]
            # if save some computation time
            if len(edges) > 1:
                edge = max(
                    edges, key=lambda e: scoring_fn.get_match_score(e, gold_edge)
                )
            else:
                edge = edges[0]
            score += scoring_fn.get_match_score(edge, gold_edge)
        else:
            free_prime_edge_index = set(range(len(edges)))
            for gold_edge in x_pairs[so]:
                edge_idx = max(
                    free_prime_edge_index,
                    key=lambda idx: scoring_fn.get_match_score(edges[idx], gold_edge),
                )
                free_prime_edge_index.remove(edge_idx)
                score += scoring_fn.get_match_score(edges[edge_idx], gold_edge)

    return score


def precision_recall_f1(
    gold_sm: "SemanticModel",
    pred_sm: "SemanticModel",
    scoring_fn: Optional[ScoringFn] = None,
    debug_dir: Optional[str] = None,
    soft_permutation_threshold: int = 50000,
    hard_permutation_threshold: int = 1000000,
) -> SmPrecisionRecallF1Output:
    if scoring_fn is None:
        scoring_fn = ScoringFn()
    pair_groups: List[PairLabelGroup] = prepare_args(gold_sm, pred_sm)

    mapping = []
    map_groups: List[PairLabelGroup] = []
    for pair in pair_groups:
        X, X_prime = pair.X, pair.X_prime

        if max(X.size, X_prime.size) == 1:
            x_prime = None if X_prime.size == 0 else X_prime.nodes[0].id
            x = None if X.size == 0 else X.nodes[0].id
            mapping.append((x_prime, x))
        else:
            map_groups.append(pair)

    bijection = PartialBijection.construct_from_mapping(mapping)
    list_of_dependent_groups: List[DependentGroups] = split_by_dependency(
        map_groups, bijection
    )

    best_bijections = []
    n_permutations = sum(
        [
            dependent_groups.get_n_permutations()
            for dependent_groups in list_of_dependent_groups
        ]
    )

    # TODO: remove debugging code or change to logging
    if n_permutations > soft_permutation_threshold:
        logger.info("Number of permutation is: {}", n_permutations)

    if n_permutations > hard_permutation_threshold:
        if debug_dir is not None:
            gold_sm.draw(os.path.join(debug_dir, "gold.png"))
            pred_sm.draw(os.path.join(debug_dir, "pred.png"))
        logger.error(
            "Permutation explosion: got {} combinations from {} pair groups",
            n_permutations,
            len(list_of_dependent_groups),
        )
        for dependent_groups in list_of_dependent_groups:
            logger.info("- {}", dependent_groups.pair_groups)
        raise PermutationExplosion(
            f"Cannot run evaluation because number of permutation is too high: {n_permutations}."
        )

    for dependent_groups in list_of_dependent_groups:
        best_bijections.append(find_best_map(dependent_groups, bijection, scoring_fn))

    for best_bijection in best_bijections:
        bijection = bijection.extends(best_bijection)

    all_groups = DependentGroups(pair_groups)

    TP = eval_score(all_groups, bijection, scoring_fn)

    if len(all_groups.X_triples) == 0:
        # gold is empty, recall must be 1
        recall = 1
    else:
        recall = TP / len(all_groups.X_triples)

    if len(all_groups.X_prime_triples) == 0:
        # predict nothing, the precision must be 1
        precision = 1
    else:
        precision = TP / len(all_groups.X_prime_triples)

    if precision == 0 and recall == 0:
        f1 = 0.0
    else:
        f1 = 2 * precision * recall / (precision + recall)

    return SmPrecisionRecallF1Output(
        f1=f1,
        precision=precision,
        recall=recall,
        bijection=bijection,
        n_corrects=TP,
        n_examples=len(all_groups.X_triples),
        n_predictions=len(all_groups.X_prime_triples),
        gold_triples=all_groups.X_triples,
        pred_triples=all_groups.X_prime_triples,
    )



from __future__ import annotations

from copy import copy
from typing import Callable, Literal, Mapping, Optional, Sequence

from graph.interface import NodeID
from rdflib import RDFS
from sm.namespaces.namespace import KnowledgeGraphNamespace
from sm.outputs.semantic_model import (
    ClassNode,
    DataNode,
    Edge,
    LiteralNode,
    LiteralNodeDataType,
    SemanticModel,
)
from sm.typing import IRI, InternalID


def replace_class_nodes_by_subject_columns(sm: SemanticModel, id_props: set[IRI]):
    """Replace a class node by its subject column (if have). One class node must have maximum one subject column"""
    rm_nodes = []
    for cnode in sm.iter_nodes():
        if isinstance(cnode, ClassNode):
            inedges = sm.in_edges(cnode.id)
            outedges = sm.out_edges(cnode.id)
            id_edges = [outedge for outedge in outedges if outedge.abs_uri in id_props]
            if len(id_edges) == 0:
                continue
            if len(id_edges) > 1:
                raise Exception(
                    f"Assuming one class node only has one subject column. Node: {cnode.id} have {len(id_edges)} subject columns: {id_edges}"
                )

            id_edge = id_edges[0]

            # update edges
            for inedge in inedges:
                sm.remove_edge(inedge.id)
                inedge.target = id_edge.target
                sm.add_edge(inedge)
            for outedge in outedges:
                sm.remove_edge(outedge.id)
                outedge.source = id_edge.target
                sm.add_edge(outedge)
            sm.remove_edge(id_edge.id)
            rm_nodes.append(cnode.id)
    for uid in rm_nodes:
        sm.remove_node(uid)


def remove_isolated_nodes(sm: SemanticModel) -> None:
    """In-placed remove isolated nodes in a semantic model"""
    rm_nodes = []
    for n in sm.iter_nodes():
        if sm.degree(n.id) == 0:
            rm_nodes.append(n.id)
    for uid in rm_nodes:
        sm.remove_node(uid)


def remove_literal_nodes(sm: SemanticModel):
    for n in sm.iter_nodes():
        if isinstance(n, LiteralNode):
            sm.remove_node(n.id)


def create_sm_nodes(
    columns: Sequence[str] | Mapping[int, str],
) -> Mapping[int, ClassNode | DataNode | LiteralNode]:
    """Create a mapping from column index to DataNode from the given columns"""
    nodes = {}
    if isinstance(columns, Sequence):
        for ci, cname in enumerate(columns):
            nodes[ci] = DataNode(ci, cname)
    else:
        for ci, cname in columns.items():
            nodes[ci] = DataNode(ci, cname)
    return nodes


def create_sm_from_cta_cpa(
    kgns: KnowledgeGraphNamespace,
    nodes: Mapping[NodeID, ClassNode | DataNode | LiteralNode],
    cpa: Sequence[tuple[NodeID, NodeID, InternalID]],
    cta: Mapping[int, InternalID],
    validate: bool = True,
    on_untype_source_column_node: Literal[
        "create-class", "remove-link"
    ] = "create-class",
    get_cls_label: Optional[Callable[[InternalID], str]] = None,
    get_prop_label: Optional[Callable[[InternalID], str]] = None,
) -> SemanticModel:
    """Create a semantic model from outputs of CPA and CTA tasks

    # Arguments
        nodes: a mapping from node id to DataNode or LiteralNode (leaf nodes) or Statement (represent n-ary relation)
            that we can't generate automatically from the cta (sources and targets in CPA should all be in this map).
        cpa: list of triples (source, target, predicate)
        cta: mapping from column index to class id
        validate: whether to validate the input to make sure it's correct classes and properties
    """
    if get_cls_label is None:
        get_cls_label = lambda ent_id: kgns.get_rel_uri(kgns.id_to_uri(ent_id))
    if get_prop_label is None:
        get_prop_label = lambda prop_id: kgns.get_rel_uri(kgns.id_to_uri(prop_id))

    if validate:
        for source, target, predicate in cpa:
            assert kgns.is_id(predicate)
        for col_index, ent_id in cta.items():
            assert kgns.is_id(ent_id)
        for node in nodes.values():
            if isinstance(node, ClassNode):
                assert kgns.is_uri_in_ns(node.abs_uri)
            elif (
                isinstance(node, LiteralNode)
                and node.datatype == LiteralNodeDataType.Entity
            ):
                assert kgns.is_uri(node.value) and kgns.is_uri_in_main_ns(node.value)

    sm = SemanticModel()
    # make a copy because when we add node into sm, we will modify the node id
    nodes = {uid: copy(u) for uid, u in nodes.items()}

    classmap = {}  # mapping from column to its class node
    # mapping from node id to sm node id, this mapping is built dynamically when iterating over cpa result.
    nodemap: dict[NodeID, int] = {}

    col2id = {u.col_index: uid for uid, u in nodes.items() if isinstance(u, DataNode)}
    for col_index, ent_id in cta.items():

        # somehow, they may end-up predict multiple classes, we need to select one
        if ent_id.find(" ") != -1:
            ent_id = ent_id.split(" ")[0]
        curl = kgns.id_to_uri(ent_id)

        cnode = ClassNode(
            abs_uri=curl,
            rel_uri=kgns.get_rel_uri(kgns.id_to_uri(ent_id)),
            readable_label=get_cls_label(ent_id),
        )
        dnode = nodes[col2id[col_index]]
        sm.add_node(cnode)
        nodemap[col2id[col_index]] = sm.add_node(dnode)
        classmap[col_index] = cnode.id
        sm.add_edge(
            Edge(
                source=cnode.id,
                target=dnode.id,
                abs_uri=str(RDFS.label),
                rel_uri=kgns.get_rel_uri(RDFS.label),
            )
        )

    for source, target, predicate in cpa:
        unode = nodes[source]
        vnode = nodes[target]

        if source not in nodemap:
            nodemap[source] = sm.add_node(unode)
        if target not in nodemap:
            nodemap[target] = sm.add_node(vnode)

    # detect and handle untype source column node
    remove_rel_from_nodes = set()

    for source, target, predicate in cpa:
        unode = nodes[source]
        if isinstance(unode, DataNode):
            if unode.col_index not in classmap:
                # discover untyped source column node
                if on_untype_source_column_node == "create-class":
                    # this data node has an outgoing edge, but it's untyped
                    # so we create an entity class to represent its type
                    cnode_id = sm.add_node(
                        ClassNode(
                            abs_uri=kgns.entity_uri,
                            rel_uri=kgns.get_rel_uri(kgns.entity_uri),
                            readable_label=kgns.entity_label,
                        )
                    )
                    classmap[unode.col_index] = cnode_id
                    sm.add_edge(
                        Edge(
                            source=cnode_id,
                            target=unode.id,
                            abs_uri=str(RDFS.label),
                            rel_uri=kgns.get_rel_uri(RDFS.label),
                        )
                    )
                else:
                    assert on_untype_source_column_node == "remove-link"
                    vnode = nodes[target]

                    if (
                        isinstance(vnode, ClassNode)
                        and vnode.abs_uri == kgns.statement_uri
                    ):
                        # this is a statement node, so we need to remove link
                        # from the statement node too because statement node
                        # can't live without the source
                        remove_rel_from_nodes.add(target)
                    remove_rel_from_nodes.add(source)

    if len(remove_rel_from_nodes) > 0:
        assert on_untype_source_column_node == "remove-link"
        cpa = [
            (source, target, predicate)
            for source, target, predicate in cpa
            if source not in remove_rel_from_nodes
        ]

    for source, target, predicate in cpa:
        unode = nodes[source]
        vnode = nodes[target]

        # if source not in nodemap:
        #     nodemap[source] = sm.add_node(unode)
        # if target not in nodemap:
        #     nodemap[target] = sm.add_node(vnode)

        if isinstance(unode, DataNode):
            # outgoing edge is from a class node instead of a data node
            assert unode.col_index in classmap
            # if unode.col_index not in classmap:
            #     # this data node has an outgoing edge, but it's untyped
            #     # so we create an entity class to represent its type
            #     curl = kgns.get_entity_abs_uri(self.ENTITY_ID)
            #     cnode_id = sm.add_node(
            #         ClassNode(
            #             abs_uri=curl,
            #             rel_uri=kgns.get_entity_rel_uri(self.ENTITY_ID),
            #             readable_label=self.kgns.entity_label,
            #         )
            #     )
            #     classmap[unode.col_index] = cnode_id
            #     sm.add_edge(
            #         Edge(
            #             source=cnode_id,
            #             target=unode.id,
            #             abs_uri=str(RDFS.label),
            #             rel_uri=kgns.get_rel_uri(RDFS.label),
            #         )
            #     )
            suid = classmap[unode.col_index]
            source = sm.get_node(suid)
        else:
            source = unode

        if isinstance(vnode, DataNode):
            # if this is an entity column, the link should map to its class
            if vnode.col_index in classmap:
                target = sm.get_node(classmap[vnode.col_index])
            else:
                target = vnode
        else:
            target = vnode

        sm.add_edge(
            Edge(
                source=source.id,
                target=target.id,
                abs_uri=(tmp_abs_uri := kgns.id_to_uri(predicate)),
                rel_uri=kgns.get_rel_uri(tmp_abs_uri),
                readable_label=get_prop_label(predicate),
            )
        )

    return sm



from __future__ import annotations

import functools
import pickle
import sqlite3
import sys
import time
from contextlib import contextmanager
from dataclasses import dataclass, field
from email.policy import default
from pathlib import Path
from traceback import format_exception
from typing import (
    Any,
    Callable,
    List,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import httpx
import numpy as np
import orjson
from loguru import logger
from sm.misc.funcs import get_classpath, import_attr
from tqdm.auto import tqdm

try:
    import ray

    has_ray = True
except ImportError:
    has_ray = False

try:
    from starlette.requests import Request
    from starlette.responses import Response
except ImportError:
    ...


@dataclass
class RayScope:
    allow_parallel: bool
    actors: dict[str, list["ray.ObjectRef"]] = field(default_factory=dict)

    def kill_actors(self, ns: Optional[str] = None):
        if ns is None:
            for actors in self.actors.values():
                for actor in actors:
                    ray.kill(actor)
            self.actors = {}
        else:
            for actor in self.actors.get(ns, []):
                ray.kill(actor)
            self.actors.pop(ns, None)

    def num_actors(self) -> int:
        return sum(len(actors) for actors in self.actors.values())


class RemoteService:
    def __init__(self, constructor: Union[type, Callable], args: tuple):
        self.object = constructor(*args)
        self.classpath = get_classpath(self.object.__class__)
        self.classargs = args

    async def __call__(self, req: "Request") -> dict | tuple | "Response":
        req = pickle.loads(await req.body())
        if req["method"] == "__meta__":
            return self.classpath, self.classargs
        try:
            return {
                "return": getattr(self.object, req["method"])(
                    *req.get("args", tuple()), **req.get("kwargs", {})
                )
            }
        except Exception as e:
            return Response(
                orjson.dumps(
                    {
                        "exception": get_classpath(e.__class__),
                        "message": str(e),
                        "stack_trace": "".join(format_exception(e)),
                    }
                ).decode(),
                media_type="application/json",
                status_code=500,
            )

    @staticmethod
    def start(
        clz: Union[type, Callable],
        args: tuple,
        options: Optional[dict] = None,
        name: Optional[str] = None,
        blocking: bool = False,
    ):
        """Deploy a service on Ray cluster.

        Args:
            clz: constructor
            args: arguments to the constructor
            options: options to deploy the service
            name: name of the service
            blocking: whether to block the main thread so users can run Ctrl+C to stop the service
        """
        from ray import serve

        serve.run(
            serve.deployment(name=name or clz.__name__, **(options or {}))(
                RemoteService
            ).bind(constructor=clz, args=args),
            blocking=blocking,
        )


class RemoteClient:
    @dataclass
    class RPC:
        slf: "RemoteClient"
        name: str

        def __call__(self, *args, **kwargs) -> Any:
            r = httpx.post(
                self.slf.endpoint,
                content=pickle.dumps(
                    {
                        "method": self.name,
                        "args": args,
                        "kwargs": kwargs,
                    }
                ),
                timeout=None,
            )
            if r.status_code != 200:
                try:
                    data = r.json()
                    exception_cls = import_attr(data["exception"])
                    exception = exception_cls(
                        data["message"] + "\n.Stack trace:\n" + data["stack_trace"]
                    )
                except:
                    logger.error("Failed to reconstruct the exception", exc_info=True)
                    raise Exception(
                        f"(Check log for more info) Failed to call {self.name}. Response ({r.status_code}), reason: {r.text}"
                    )
                raise exception

            return r.json()["return"]

    def __init__(self, cls: type, args: tuple, endpoint: str):
        self.cls = cls
        self.args = args
        self.endpoint = endpoint
        self.is_validated = False

    def __getattr__(self, name: str) -> Any:
        if not self.is_validated:
            # only validate once when we actually use the service
            # if the remote service is correct...
            classpath, classargs = (
                httpx.post(
                    self.endpoint,
                    content=pickle.dumps({"method": "__meta__"}),
                    timeout=None,
                )
                .raise_for_status()
                .json()
            )
            if not (
                classpath == get_classpath(self.cls) and tuple(classargs) == self.args
            ):
                raise ValueError(
                    f"Remote service is not correct. Get ({classpath})({classargs}) instead of ({get_classpath(self.cls)})({self.args})"
                )
            self.is_validated = True

        return RemoteClient.RPC(self, name)


R = TypeVar("R")
OBJECTS = {}
ray_initargs: dict = {}
ray_scopes: list[RayScope] = []


def require_ray(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not has_ray:
            raise ImportError("ray is required for function: %s" % func.__name__)
        return func(*args, **kwargs)

    return wrapper


def is_parallelizable() -> bool:
    """Check whether the current context is parallelable."""
    global ray_scopes
    return len(ray_scopes) > 0 and ray_scopes[-1].allow_parallel


@contextmanager
def allow_parallel():
    """Mark this code block as parallelizable."""
    global ray_scopes

    try:
        ray_scopes.append(RayScope(allow_parallel=True))
        yield
    finally:
        scope = ray_scopes.pop()
        scope.kill_actors()


@contextmanager
def no_parallel():
    """Mark this code block as non-parallelizable."""
    global ray_scopes

    try:
        ray_scopes.append(RayScope(allow_parallel=False))
        yield
    finally:
        scope = ray_scopes.pop()
        assert scope.num_actors() == 0


def set_ray_init_args(**kwargs):
    global ray_initargs
    ray_initargs = kwargs


@require_ray
def add_ray_actors(
    cls: type[R],
    args: tuple,
    ns: str,
    size: int = 1,
    scope: int = -1,
    remote_options: Optional[dict] = None,
) -> list["ray.ObjectRef[R]"]:
    """Create a ray actor and return the actor ref. Also, store the actor ref in an scope identified by the index.

    Args:
        cls: the class of the actor.
        args: the arguments to create the actor.
        size: number of actors to create
        scope: scope index
    """
    global ray_scopes
    scope_data = ray_scopes[scope]
    assert scope_data.allow_parallel, "Cannot create actors in a non-parallel context"
    if ns not in scope_data.actors:
        scope_data.actors[ns] = []

    if len(scope_data.actors[ns]) < size and remote_options is not None:
        wrapper = ray.remote(**remote_options)
    else:
        wrapper = ray.remote

    while len(scope_data.actors[ns]) < size:
        scope_data.actors[ns].append(wrapper(cls).remote(*args))
    return scope_data.actors[ns]


def get_ray_actors(ns: str, scope: int = -1) -> list["ray.ObjectRef"]:
    global ray_scopes
    scope_data = ray_scopes[scope]
    assert scope_data.allow_parallel, "Cannot create actors in a non-parallel context"
    return scope_data.actors.get(ns, [])


@require_ray
def ray_init(**kwargs):
    if not ray.is_initialized():
        logger.info("Initialize ray with args: {}", kwargs)
        ray.init(**kwargs)


@overload
def ray_put(val: R, using_ray: Literal[True] = True) -> "ray.ObjectRef[R]": ...


@overload
def ray_put(val: R, using_ray: Literal[False]) -> R: ...


@overload
def ray_put(val: R, using_ray: bool) -> Union["ray.ObjectRef[R]", R]: ...


@require_ray
def ray_put(val: R, using_ray: bool = True) -> Union["ray.ObjectRef[R]", R]:
    global ray_initargs
    if not using_ray:
        return val
    ray_init(**ray_initargs)
    return ray.put(val)


def ray_get(val: "ray.ObjectRef[R]") -> R:
    obj = ray.get(val)
    if isinstance(obj, np.ndarray):
        return np.copy(obj)  # type: ignore
    return obj


@require_ray
def ray_shutdown():
    ray.shutdown()


@require_ray
def ray_get_num_gpu() -> float:
    ray_init(**ray_initargs)
    return ray.available_resources().get("GPU", 0)


@require_ray
def ray_map(
    fn: Union[Callable[..., "ray.ObjectRef[R]"], Callable[..., R]],
    args_lst: Sequence[Sequence],
    verbose: bool = False,
    poll_interval: float = 0.1,
    concurrent_submissions: int = 3000,
    desc: Optional[str] = None,
    using_ray: bool = True,
    is_func_remote: bool = True,
    remote_options: Optional[dict] = None,
    before_shutdown: Optional[Callable[[Any], Any]] = None,
    auto_shutdown: bool = False,
) -> List[R]:
    """
    Args:
        before_shutdown: if you use numpy arrays, shutdown ray cluster will released the shared memory and thus, may corrupt the arrays later. You should use
            before_shutdown to copy the data before shutdown. This only applies to the case where using_ray=True && auto_shutdown=True.
    """
    global ray_initargs

    if not using_ray:
        # run locally without ray, usually for debugging
        if is_func_remote:
            localfn: Callable[..., R] = fn.__wrapped__
        else:
            localfn = cast(Callable[..., R], fn)
        output = []
        for arg in tqdm(args_lst, desc=desc, disable=not verbose or len(args_lst) <= 1):
            newarg = []
            for x in arg:
                if isinstance(x, ray.ObjectRef):
                    newarg.append(ray_get(x))
                else:
                    newarg.append(x)
            try:
                output.append(localfn(*newarg))
            except:
                logger.error("ray_map failed at item index {}", len(output))
                raise
        return output

    ray_init(**ray_initargs)

    n_jobs = len(args_lst)

    if is_func_remote:
        remote_fn = cast(Callable[..., "ray.ObjectRef[R]"], fn)
    else:
        if remote_options is None:
            wrapper = ray.remote
        else:
            wrapper = ray.remote(**remote_options)
        remote_fn: Callable[..., "ray.ObjectRef[R]"] = wrapper(fn).remote  # type: ignore

    with tqdm(total=n_jobs, desc=desc, disable=not verbose) as pbar:
        output: List[R] = [None] * n_jobs  # type: ignore

        notready_refs = []
        ref2index = {}
        for i, args in enumerate(args_lst):
            # submit a task and add it to not ready queue and ref2index
            ref = remote_fn(*args)
            notready_refs.append(ref)
            ref2index[ref] = i

            # when the not ready queue is full, wait for some tasks to finish
            while len(notready_refs) >= concurrent_submissions:
                ready_refs, notready_refs = ray.wait(
                    notready_refs, timeout=poll_interval
                )
                pbar.update(len(ready_refs))
                for ref in ready_refs:
                    try:
                        output[ref2index[ref]] = ray_get(ref)
                    except:
                        logger.error("ray_map failed at item index {}", ref2index[ref])
                        raise

        while len(notready_refs) > 0:
            ready_refs, notready_refs = ray.wait(notready_refs, timeout=poll_interval)
            pbar.update(len(ready_refs))
            for ref in ready_refs:
                try:
                    output[ref2index[ref]] = ray_get(ref)
                except:
                    logger.error("ray_map failed at item index {}", ref2index[ref])
                    raise

        if auto_shutdown:
            if before_shutdown is not None:
                output = [before_shutdown(x) for x in output]
            ray_shutdown()
        return output


@require_ray
def ray_actor_map(
    actor_fns: list[Callable],
    args_lst: Sequence[Sequence],
    verbose: bool = False,
    poll_interval: float = 0.1,
    concurrent_submissions: int = 3000,
    desc: Optional[str] = None,
    postprocess: Optional[Callable[[Any], Any]] = None,
    before_shutdown: Optional[Callable[[Any], Any]] = None,
    auto_shutdown: bool = False,
):
    n_jobs = len(args_lst)
    with tqdm(total=n_jobs, desc=desc, disable=not verbose) as pbar:
        output: list = [None] * n_jobs

        notready_refs = []
        ref2index = {}
        for i, args in enumerate(args_lst):
            # submit a task and add it to not ready queue and ref2index
            ref = actor_fns[i % len(actor_fns)](*args)
            notready_refs.append(ref)
            ref2index[ref] = i

            # when the not ready queue is full, wait for some tasks to finish
            while len(notready_refs) >= concurrent_submissions:
                ready_refs, notready_refs = ray.wait(
                    notready_refs, timeout=poll_interval
                )
                pbar.update(len(ready_refs))
                for ref in ready_refs:
                    try:
                        output[ref2index[ref]] = ray_get(ref)
                    except:
                        logger.error(
                            "ray_actor_map failed at item index {}", ref2index[ref]
                        )
                        raise

        while len(notready_refs) > 0:
            ready_refs, notready_refs = ray.wait(notready_refs, timeout=poll_interval)
            pbar.update(len(ready_refs))
            for ref in ready_refs:
                try:
                    output[ref2index[ref]] = ray_get(ref)
                except:
                    logger.error(
                        "ray_actor_map failed at item index {}", ref2index[ref]
                    )
                    raise

        if postprocess is not None:
            output = [postprocess(x) for x in output]

        if auto_shutdown:
            if before_shutdown is not None:
                output = [before_shutdown(x) for x in output]
            ray_shutdown()

        return output


def enhance_error_info(getid: Union[Callable[..., str], str]):
    """Enhancing error report by printing out tracable id of the input arguments.

    Args:
        getid: a function that takes the same arguments as the wrapped function and return a tracable id.
            If msg is a string, it is a list of accessors joined by dot. Each accessor is either a number
            (to call __getitem__) or a string (to call __getattr__). The first accessor is always the number
            which is the argument index that will be used to extract the traceable id from. For example: 0.table.table_id
    """

    if isinstance(getid, str):

        def get_id_fn(*args, **kwargs):
            assert len(kwargs) == 0
            ptr = args
            for accessor in getid.split("."):
                if accessor.isdigit():
                    ptr = ptr[int(accessor)]
                else:
                    ptr = getattr(ptr, accessor)
            return ptr

    else:
        get_id_fn = getid

    def wrap_func(func):
        func_name = func.__name__

        @functools.wraps(func)
        def fn(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except BaseException as e:
                if hasattr(sys, "gettrace") and sys.gettrace() is not None:
                    # for debug mode in vscode...
                    logger.error(
                        f"Failed to run {func_name} with {get_id_fn(*args, **kwargs)}"
                    )
                    raise
                else:
                    raise Exception(
                        f"Failed to run {func_name} with {get_id_fn(*args, **kwargs)}"
                    ) from e

        return fn

    return wrap_func


def track_runtime(getid: Union[Callable[..., str], str], outfile: Union[str, Path]):
    if isinstance(getid, str):

        def get_id_fn(*args, **kwargs):
            assert len(kwargs) == 0
            ptr = args
            for accessor in getid.split("."):
                if accessor.isdigit():
                    ptr = ptr[int(accessor)]
                else:
                    ptr = getattr(ptr, accessor)
            return ptr

    else:
        get_id_fn = getid

    init_db = not Path(outfile).exists()
    db = sqlite3.connect(outfile)
    if init_db:
        with db:
            db.execute("CREATE TABLE timesheet(func TEXT, name TEXT, time REAL)")

    def wrap_func(func):
        func_name = func.__name__

        @functools.wraps(func)
        def fn(*args, **kwargs):
            start = time.time()
            try:
                return func(*args, **kwargs)
            finally:
                end = time.time()
                with db:
                    db.execute(
                        "INSERT INTO timesheet VALUES (:func, :name, :time)",
                        {
                            "func": func_name,
                            "name": get_id_fn(*args, **kwargs),
                            "time": end - start,
                        },
                    )

        return fn

    return wrap_func


def get_instance(constructor: Callable[[], R], name: Optional[str] = None) -> R:
    """A utility function to get a singleton, which can be created from the given constructor.

    One use case of this function is we have a big object that is expensive to send
    to individual task repeatedly. If the process are retrieved from a pool,
    this allows us to create the object per process instead of per task.
    """
    global OBJECTS

    if name is None:
        assert (
            constructor.__name__ != "<lambda>"
        ), "Cannot use lambda as a name because it will keep changing"
        name = constructor  # type: ignore

    if name not in OBJECTS:
        logger.trace("Create a new instance of {}", name)
        OBJECTS[name] = constructor()
    return OBJECTS[name]



from __future__ import annotations

import glob
import importlib
import math
import re
from contextlib import contextmanager
from inspect import signature
from multiprocessing import get_context
from multiprocessing.pool import ThreadPool
from operator import itemgetter
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    KeysView,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from loguru import logger
from tqdm.auto import tqdm
from typing_extensions import TypeGuard

K = TypeVar("K")
V = TypeVar("V")
TYPE_ALIASES = {"typing.List": "list", "typing.Dict": "dict", "typing.Set": "set"}


def str2bool(x):
    assert x in {"True", "False", "true", "false", "null"}
    if x == "null":
        return None
    return x.lower() == "true"


def nullable_str(x):
    if x == "null":
        return None
    return x


def str2int(x):
    if x == "null":
        return None
    return int(x)


def assert_not_null(x: Optional[V]) -> V:
    assert x is not None
    return x


def assert_all_item_not_null(lst: list[Optional[V]]) -> list[V]:
    assert all(v is not None for v in lst)
    return lst  # type: ignore


def assert_one_item(lst: list[V]) -> V:
    assert len(lst) == 1
    return lst[0]


def assert_not_empty(lst: list[V]) -> list[V]:
    assert len(lst) > 0
    return lst


def assert_isinstance(x: Any, cls: type[V]) -> V:
    if not isinstance(x, cls):
        raise Exception(f"{type(x)} doesn't match with {type(cls)}")
    return x


def assert_is_unique(lst: list[V]) -> bool:
    return len(set(lst)) == len(lst)


def assert_desc_sorted(lst: list[V], value: Callable[[V], float | int]):
    if len(lst) == 0:
        return lst

    v0 = value(lst[0])
    for i in range(1, len(lst)):
        vi = value(lst[i])
        assert v0 >= vi, (i, v0, vi)
        v0 = vi

    return lst


def get_max_sorted_desc(lst: list[V], value: Callable[[V], float | int]) -> V:
    assert len(lst) >= 1
    v0 = value(lst[0])
    assert all(v0 > value(x) for x in lst[1:])
    return lst[0]


def is_not_null(x: Optional[V]) -> TypeGuard[V]:
    return x is not None


def identity_func(x):
    return x


def percentage(
    a: Union[float, int],
    b: Union[float, int],
    format: Optional[Callable[[Union[float, int]], str]] = None,
) -> str:
    percent = a * 100 / (b if b > 0 else 0.01)
    if format is None:
        return "%.2f%% (%d/%d)" % (percent, a, b)
    return "%.2f%% (%s/%s)" % (percent, format(a), format(b))


def is_non_decreasing_sequence(
    lst: Union[list[Union[int, float]], list[int], list[float]]
) -> bool:
    return len(lst) == 0 or (all(lst[i - 1] <= lst[i] for i in range(1, len(lst))))


def is_monotonic_decreasing(lst: Sequence[Union[int, float]]) -> bool:
    return len(lst) == 0 or (all(lst[i - 1] >= lst[i] for i in range(1, len(lst))))


def filter_duplication(
    lst: Iterable[V], key_fn: Optional[Callable[[V], Any]] = None
) -> list[V]:
    keys = set()
    new_lst = []
    if key_fn is not None:
        for item in lst:
            k = key_fn(item)
            if k in keys:
                continue

            keys.add(k)
            new_lst.append(item)
    else:
        for k in lst:
            if k in keys:
                continue
            keys.add(k)
            new_lst.append(k)
    return new_lst


def get_latest_version(file_pattern: Union[str, Path]) -> int:
    """Assuming the file pattern select list of files tagged with an integer version for every run, this
    function return the latest version number that you can use to name your next run.

    For example:
    1. If your pattern matches folders: version_1, version_5, version_6, this function will return 6.
    2. If your pattern does not match anything, return 0
    """
    files = [Path(file) for file in sorted(glob.glob(str(file_pattern)))]
    if len(files) == 0:
        return 0

    versions: list[int] = []
    for file in files:
        match = re.match(r"[^0-9]*(\d+)[^0-9]*", file.name)
        if match is None:
            raise Exception("Invalid naming")
        versions.append(int(match.group(1)))

    return sorted(versions)[-1]


def get_incremental_path(
    path: Union[str, Path],
    create_if_missing: bool = True,
    delimiter_char: Optional[str] = None,
) -> Path:
    """Get an incremental path.

    Example:
        >>> get_incremental_path("/test/data")
        '/test/data_01'
        >>> mkdir('/test/data_01')
        >>> # create a new folder with version 01, so the next time you call, it will be `data_02`
        >>> get_incremental_path("/test/data")
        '/test/data_02'

    Arguments:
        path: path to the folder you want to get incremental path
        create_if_missing:
            if the path has a suffix, it is treated as a file and if its parent directory does not exist, create the parent directory.
            if the path does not have a suffix, it is treated as a folder and create if it does not exist.
        delimiter_char: if you want to use a different delimiter, you can specify it here.
            Default it is _ for no suffix path (folder) and . for file
    """
    path = Path(str(path))
    if delimiter_char is None:
        if path.suffix == "":
            delimiter_char = "_"
        else:
            delimiter_char = "."

    pattern = path.parent / f"{path.stem}{delimiter_char}*{path.suffix}"
    version = get_latest_version(pattern) + 1

    newpath = path.parent / f"{path.stem}{delimiter_char}{version:02d}{path.suffix}"
    if create_if_missing:
        if path.suffix == "":
            newpath.mkdir(parents=True)
        else:
            newpath.parent.mkdir(parents=True, exist_ok=True)
    return newpath


def get_latest_path(path: Union[str, Path]) -> Optional[Path]:
    path = Path(str(path))
    pattern = path.parent / f"{path.stem}*{path.suffix}"
    version = get_latest_version(pattern)
    if version == 0:
        return None

    (match_path,) = list(path.parent.glob(f"{path.stem}*{version:02d}{path.suffix}"))
    return match_path


def auto_wrap(
    word: str,
    max_char_per_line: int,
    delimiters: Optional[list[str]] = None,
    camelcase_split: bool = True,
) -> str:
    """
    Treat this as optimization problem, where we trying to minimize the number of line break
    but also maximize the readability in each line, i.e: maximize the number of characters in each lines

    Using greedy search.
    :param word:
    :param max_char_per_line:
    :param delimiters:
    :return:
    """
    # split original word by the delimiters
    if delimiters is None:
        delimiters = [" ", ":", "_", "/"]

    sublines: list[str] = [""]
    for i, c in enumerate(word):
        if c not in delimiters:
            sublines[-1] += c

            if (
                camelcase_split
                and not c.isupper()
                and i + 1 < len(word)
                and word[i + 1].isupper()
            ):
                # camelcase_split
                sublines.append("")
        else:
            sublines[-1] += c
            sublines.append("")

    new_sublines: list[str] = [""]
    for line in sublines:
        if len(new_sublines[-1]) + len(line) <= max_char_per_line:
            new_sublines[-1] += line
        else:
            new_sublines.append(line)

    return "\n".join(new_sublines)


def exchange_keyvalue(
    odict: dict[K, V] | dict[K, Iterable[V]], is_bijection: bool = False
) -> dict[V, list[K]] | dict[V, K]:
    out = {}
    if is_bijection:
        for k, v in odict.items():
            assert v not in out
            out[v] = k
    else:
        for k, v in odict.items():
            if isinstance(v, Iterable):
                for x in v:
                    if x not in out:
                        out[x] = []
                    out[x].append(k)
            else:
                if v not in out:
                    out[v] = []
                out[v].append(k)
    return out


def flatten_dict(odict: dict, result: Optional[dict] = None, prefix: str = ""):
    if result is None:
        result = {}

    for k, v in odict.items():
        if isinstance(v, dict):
            flatten_dict(v, result, prefix=prefix + k + ".")
        else:
            result[prefix + k] = v
    return result


def flatten_list(lst: list) -> list:
    """Flatten nested list, anything that is instance of a list get flatten"""
    output = []
    for item in lst:
        if isinstance(item, list):
            for subitem in item:
                if isinstance(subitem, list):
                    output += flatten_list(subitem)
                else:
                    output.append(subitem)
        else:
            output.append(item)
    return output


def batch(size: int, *vars, return_tuple: bool = False):
    """Batch the variables into batches of size. When vars is a single variable,
    it will return a list of batched values instead of list of tuple of batched values.

    If we want to batch a single variable to a list of tuple of batched values, set
    return_tuple to True.
    """
    output = []
    n = len(vars[0])
    if len(vars) == 1 and not return_tuple:
        for i in range(0, n, size):
            output.append(vars[0][i : i + size])
    else:
        for i in range(0, n, size):
            output.append(tuple(var[i : i + size] for var in vars))
    return output


def group_by(lst: Iterable[V], key: Callable[[V], K]) -> dict[K, list[V]]:
    odict = {}
    for item in lst:
        k = key(item)
        if k not in odict:
            odict[k] = []
        odict[k].append(item)
    return odict


def create_group_by_index(
    lst: Iterable[V], key: Callable[[V], K]
) -> dict[K, list[int]]:
    odict = {}
    for i, item in enumerate(lst):
        k = key(item)
        if k not in odict:
            odict[k] = []
        odict[k].append(i)
    return odict


def cluster(same_as: Sequence[tuple[K, K]]) -> list[list[K]]:
    key2group: dict[K, int] = {}
    groups: list[set[K]] = []
    for k1, k2 in same_as:
        if k1 not in key2group and k2 not in key2group:
            groups.append({k1, k2})
            key2group[k1] = len(groups) - 1
            key2group[k2] = len(groups) - 1
        elif k1 in key2group and k2 not in key2group:
            groups[key2group[k1]].add(k2)
            key2group[k2] = key2group[k1]
        elif k2 in key2group and k1 not in key2group:
            groups[key2group[k2]].add(k1)
            key2group[k1] = key2group[k2]
        elif key2group[k1] != key2group[k2]:
            newgroup = groups[key2group[k1]].union(groups[key2group[k2]])
            groups[key2group[k1]] = set()
            groups[key2group[k2]] = set()
            groups.append(newgroup)
            for k in newgroup:
                key2group[k] = len(groups) - 1

    return [sorted(g) for g in groups if len(g) > 0]


def make_dict(iter: Iterable[V], key: Callable[[V], K]) -> dict[K, V]:
    odict = {}
    for item in iter:
        k = key(item)
        assert k not in odict
        odict[k] = item
    return odict


def datasize(num, suffix="B"):
    """Get human friendly file size
    https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597#gistcomment-2845059

    Args:
        num (int): Bytes value
        suffix (str, optional): Unit. Defaults to 'B'.

    Returns:
        str: file size0
    """
    if num == 0:
        return "0"
    magnitude = int(math.floor(math.log(num, 1024)))
    val = num / math.pow(1024, magnitude)
    if magnitude > 7:
        return "{:3.1f}{}{}".format(val, "Yi", suffix)
    return "{:3.1f}{}{}".format(
        val, ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"][magnitude], suffix
    )


class IntegerEncoder:
    def __init__(self):
        self.encoding: dict[str | tuple[str, ...], int] = {}
        self.values: list[int] = []

    def append(self, val: str | tuple[str, ...]):
        if val not in self.encoding:
            self.encoding[val] = len(self.encoding)
        self.values.append(self.encoding[val])

    def get_decoder(self) -> list:
        return get_decoder(self.encoding)


class KnownSizeIntegerEncoder:
    def __init__(self, encoder: dict[str | tuple[str, ...], int], size: int):
        self.encoder = encoder
        self.values = np.zeros(size, dtype=np.int32)

    def __setitem__(self, idx: int | slice, val: str | tuple[str, ...]):
        if val not in self.encoder:
            self.encoder[val] = len(self.encoder)
        self.values[idx] = self.encoder[val]

    def get_decoder(self) -> list:
        return get_decoder(self.encoder)


def get_encoder(decoder: list[V]) -> dict[V, int]:
    assert_is_unique(decoder)
    return {v: i for i, v in enumerate(decoder)}


def get_decoder(encoder: dict[V, int]) -> list[V]:
    counter = 0
    output = []
    for key, val in encoder.items():
        assert val == counter, (val, counter)
        counter += 1
        output.append(key)
    return output


class ParallelMapFnWrapper:
    def __init__(self, fn: Callable, ignore_error=False):
        self.fn = fn
        fn_params = signature(fn).parameters
        self.spread_fn_args = len(fn_params) > 1
        self.ignore_error = ignore_error

    def run(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, r
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise


def parallel_map(
    fn,
    inputs,
    show_progress=False,
    progress_desc="",
    is_parallel=True,
    use_threadpool=False,
    n_processes: Optional[int] = None,
    ignore_error: bool = False,
    start_method: Literal["fork", "spawn", "forkserver"] = "fork",
):
    if not is_parallel:
        iter = (fn(item) for item in inputs)
        if show_progress:
            iter = tqdm(iter, total=len(inputs), desc=progress_desc)
        return list(iter)

    if use_threadpool:
        with ThreadPool(processes=n_processes) as pool:
            iter = pool.imap_unordered(
                ParallelMapFnWrapper(fn, ignore_error).run, enumerate(inputs)
            )
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    else:
        with get_context(start_method).Pool(processes=n_processes) as pool:
            iter = pool.imap_unordered(
                ParallelMapFnWrapper(fn, ignore_error).run, enumerate(inputs)
            )
            if show_progress:
                iter = tqdm(iter, total=len(inputs), desc=progress_desc)
            results = list(iter)
            results.sort(key=itemgetter(0))
    return [v for i, v in results]


@contextmanager
def print2file(file_path: Union[str, Path], mode="w", file_only: bool = False):
    """Yield a print function that can be both print to file or print to std"""
    Path(file_path).parent.mkdir(exist_ok=True, parents=True)
    origin_print = print
    with open(str(file_path), mode) as f:

        def print_fn(*args):
            if not file_only:
                origin_print(*args)
            origin_print(*args, file=f)

        try:
            yield print_fn
        finally:
            pass


K = TypeVar("K")
V = TypeVar("V")
V2 = TypeVar("V2")


class DictProxy(Mapping[K, V2]):
    """Dictionary proxy to access objects' property

    Args:
        odict: dictionary of object
        access: function to access property of an object
    """

    def __init__(self, odict: dict[K, V], access: Callable[[V], V2]):
        self.odict = odict
        self.access = access

    def __iter__(self):
        return self.odict.__iter__()

    def __getitem__(self, item):
        return self.access(self.odict[item])

    def __contains__(self, item):
        return item in self.odict

    def keys(self) -> KeysView[K]:
        return self.odict.keys()

    def values(self):
        return (self.access(v) for v in self.odict.values())

    def items(self):
        return ((k, self.access(v)) for k, v in self.odict.items())


def get_classpath(type: Type | Callable) -> str:
    if type.__module__ == "builtins":
        return type.__qualname__

    if hasattr(type, "__qualname__"):
        return type.__module__ + "." + type.__qualname__

    # typically a class from the typing module
    if hasattr(type, "_name") and type._name is not None:
        path = type.__module__ + "." + type._name
        if path in TYPE_ALIASES:
            path = TYPE_ALIASES[path]
    elif hasattr(type, "__origin__") and hasattr(type.__origin__, "_name"):
        # found one case which is typing.Union
        path = type.__module__ + "." + type.__origin__._name
    else:
        raise NotImplementedError(type)

    return path


def import_func(func_ident: str) -> Callable:
    """Import function from string, e.g., sm.misc.funcs.import_func"""
    lst = func_ident.rsplit(".", 2)
    if len(lst) == 2:
        module, func = lst
        cls = None
    else:
        module, cls, func = lst
        try:
            importlib.import_module(module + "." + cls)
            module = module + "." + cls
            cls = None
        except ModuleNotFoundError as e:
            if e.name == (module + "." + cls):
                pass
            else:
                raise

    module = importlib.import_module(module)
    if cls is not None:
        module = getattr(module, cls)

    return getattr(module, func)


def import_attr(attr_ident: str):
    lst = attr_ident.rsplit(".", 1)
    module, cls = lst
    module = importlib.import_module(module)
    return getattr(module, cls)


class Proxy:
    """Proxy object, can be used for delayed initialization"""

    def __init__(self, _object):
        self._object = _object

    def __getattr__(self, name):
        return getattr(self._object(), name)



from __future__ import annotations

from typing import Annotated

IRI = Annotated[str, "IRI (e.g., https://www.wikidata.org/wiki/Q5)"]
RelIRI = Annotated[str, "Relative Internationalized Resource Identifier"]
InternalID = Annotated[str, "Internal ID (e.g., Q5)"]
ExampleId = Annotated[str, "Id of an example (e.g., table id)"]
ColumnIndex = Annotated[int, "Index of a column in the table"]



from .column import Column
from .context import Context
from .link import WIKIDATA_NIL_ENTITY, EntityId, EntityIdWithScore, Link
from .table import ColumnBasedTable

__all__ = [
    "Column",
    "ColumnBasedTable",
    "Context",
    "EntityId",
    "EntityIdWithScore",
    "Link",
    "WIKIDATA_NIL_ENTITY",
]
